"""
-------------------------------------------------------------------------------
   File       : test_ggg.sage
   Description: Test file for ggg.sage
   Copyright  : 2018, Robin Larrieu
-------------------------------------------------------------------------------
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
                   http://www.gnu.org/licenses/

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------
"""

load("ggg.sage")

def hline():
    print "------------------------------------------------------"

def hline2():
    print "======================================================"

def result(b):
    print "OK" if b else "KO"


#=================================
# Usual reduction algorithm
#=================================

def test_reduction(G,Q,Rem,P):
    """
    Verify that the extended reduction is correct, that is
    P = sum_i Q_i * G_i + Rem
    """
    check = Rem
    for i in range(len(G)):
        check += Q[i] * G[i]
    return P == check
    
def test_reduction_zealous(G,P):
    """
    Verify correctness of the schoolbook reduction algorithm
    """
    print "Try to reduce P mod G (schoolbook algorithm)...",
    (Q,Rem) = dichotomic_zealous_reduction(G,P)
    result(test_reduction(G,Q,Rem,P))

def test_reduction_relaxed(G,P,k):
    """
    Verify correctness of the relaxed reduction algorithm
    (assumes that the term order being used is the k-order)
    """
    print "Try to reduce P mod G (relaxed algorithm)...",
    (Q,Rem) = dichotomic_relaxed_reduction(G,P,k)
    result(test_reduction(G,Q,Rem,P))


#=================================
# Main algorithm
#=================================

def test_concise_repr(n,m,show_supports=False):
    """
    For random polynomials A,B of total degree n,m,
    compute G and its concise representation, then check
    the correctness.
    
    With show_supports=True (default False), also print
    the supports of the truncated elements
    """
    hline()
    print "Test the concise representation for random elements of degrees %d and %d" % (n,m)
    A=random_element(n); B=random_element(m)
    CR = concise_repr(A,B)
    (Gsharp, retrac) = CR
    G = [ g for g in Gsharp ]
    L = len(G)
    for i in range(L-2):
        # ( G_{i+1}, G_{i+2} ) = M_{1,i} * ( G_i, G_{i+1} )
        G[i+2] = (retrac[0][i] * vector([G[i], G[i+1]]))[1]
    print "Checking Groebner stairs"
    LM = [ g.lm() for g in Gsharp ]
    print "  ", LM,
    if LM == generic_stairs(n,m):
        print "OK"
    else:
        print "KO"
        print "  expected", generic_stairs(n,m)
        return false
    expectedGsharp = [ truncate_k_degree(G[i], 1, truncation_precision(i,n)) for i in range(L) ]
    print "\nChecking truncations",
    if show_supports:
        for g in Gsharp:
            show_support(g)
    if Gsharp == expectedGsharp:
        print "OK"
    else:
        print "KO"
        return False
    print "\nChecking higher order retraction matrices:"
    Lambda = 1; l = 2
    while Lambda < len(retrac):
        print "  Step 2^%d:" % Lambda,
        MLambda = retrac[Lambda]
        for j in range(len(MLambda)):
            i = j*l
            Giip1 = vector([G[i], G[i+1]])
            expected = vector([G[n], 0]) if j == len(MLambda)-1 else vector([G[i+l], G[i+l+1]])
            if MLambda[j] * Giip1 != expected:
                print "KO"
                return False
        print "OK"
        Lambda += 1; l *= 2
    return True


def test_grevlex_fast_reduce(G, Gsharp, retrac, P):
    """
    Verify the correctness of the fast reduction algorithm (grevlex case)
    """
    print "Reducing P mod G (deg P = %d)..." % P.degree(),
    (Q,Rem) = grevlex_fast_reduce( (Gsharp, retrac), P )
    result(test_reduction(G,Q,Rem,P))

def test_grevlex_fast_reduce_bis(A,B,P):
    """
    For A,B generic of total degree n,m, compute G and its concise representation.
    Then reduce P modulo G using the fast reduction algorithm and check the result.
    """
    (Gsharp, retrac) = concise_repr(A,B)
    G = [ g for g in Gsharp ]
    for i in range(len(G)-2):
        # ( G_{i+1}, G_{i+2} ) = M_{1,i} * ( G_i, G_{i+1} )
        G[i+2] = (retrac[0][i] * vector([G[i], G[i+1]]))[1]
    test_grevlex_fast_reduce(G, Gsharp, retrac, P)

def test_grevlex_GB(A,B):
    """
    Check that grevlex_groebner_basis (algorithm 5 in the paper)
    actually computes the reduced Groebner basis
    """
    hline()
    print "Check computation of the reduced Groebner basis"
    G1 = grevlex_groebner_basis(A,B)
    I = R.change_ring(order=grevlex).ideal(A,B)
    G2 = I.groebner_basis()
    # sort G2 by increasing x-power in the leading monomial
    G2 = sort_basis(G2)
    result(G1 == G2)

#=================================
# Modified version from section 7
#=================================

def test_grevlex_fast_reduce_var(G, AugConciseRepr, P):
    """
    Verify the correctness of the fast reduction algorithm (with modification)
    """
    print "Reducing P mod G (deg P = %d)..." % P.degree(),
    (Q,Rem) = grevlex_fast_reduce_variant( AugConciseRepr, P )
    result(test_reduction(G,Q,Rem,P))

def test_grevlex_fast_reduce_var_bis(A,B,P):
    """
    For A,B generic of total degree n,m, compute G and its concise representation.
    Then reduce P modulo G using the fast reduction algorithm and check the result.
    """
    AugConciseRepr = augmented_concise_repr(A,B)
    (Gsharp, retrac, G12, M12) = AugConciseRepr
    G = [ g for g in Gsharp ]
    for i in range(len(G)-2):
        # ( G_{i+1}, G_{i+2} ) = M_{1,i} * ( G_i, G_{i+1} )
        G[i+2] = (retrac[0][i] * vector([G[i], G[i+1]]))[1]
    test_grevlex_fast_reduce(G, AugConciseRepr, P)

#=================================
# Main tests
#=================================

def main_tests():
    print "Working on", R

    test_concise_repr(10,10)
    print
    test_concise_repr(20,30)

    print

    hline2()
    print "Testing the reduction algorithm for deg(A)=deg(B)=20"
    print
    hline()
    print "Generating random polynomials and computing the Groebner basis"
    A=random_element(20)
    B=random_element(20)
    (Gsharp, retrac) = concise_repr(A,B)
    G = [ g for g in Gsharp ]
    for i in range(len(G)-2):
        # ( G_{i+1}, G_{i+2} ) = M_{1,i} * ( G_i, G_{i+1} )
        G[i+2] = (retrac[0][i] * vector([G[i], G[i+1]]))[1]
    print
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(15)) # d < n
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(20)) # d = n = m
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(25)) # m < d < deg(G_n)
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(39)) # d = deg(G_n)
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(50)) # deg(G_n) < d
    
    hline()
    print "Checking the modified version from section 7"
    AugConciseRepr = augmented_concise_repr(A,B)
    print
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(15)) # d < n
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(20)) # d = n = m
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(25)) # m < d < deg(G_n)
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(39)) # d = deg(G_n)
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(50)) # deg(G_n) < d
    
    print
    hline2()
    print "Testing the reduction algorithm for deg(A)=20, deg(B)=30"
    print
    hline()
    print "Generating random polynomials and computing the Groebner basis"
    A=random_element(20)
    B=random_element(30)
    (Gsharp, retrac) = concise_repr(A,B)
    G = [ g for g in Gsharp ]
    for i in range(len(G)-2):
        # ( G_{i+1}, G_{i+2} ) = M_{1,i} * ( G_i, G_{i+1} )
        G[i+2] = (retrac[0][i] * vector([G[i], G[i+1]]))[1]
    print
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(15)) # d < n
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(20)) # d = n
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(25)) # n < d < m
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(30)) # d = m
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(35)) # m < d < deg(G_n)
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(49)) # d = deg(G_n)
    test_grevlex_fast_reduce(G, Gsharp, retrac, random_element(60)) # deg(G_n) < d
    
    hline()
    print "Checking the modified version from section 7"
    AugConciseRepr = augmented_concise_repr(A,B)
    print
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(15)) # d < n
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(20)) # d = n
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(25)) # n < d < m
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(30)) # d = m
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(35)) # m < d < deg(G_n)
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(49)) # d = deg(G_n)
    test_grevlex_fast_reduce_var(G, AugConciseRepr, random_element(60)) # deg(G_n) < d
    
    
    
    print
    test_grevlex_GB(A,B)

main_tests()

