"""
-------------------------------------------------------------------------------
   File       : test_redbivar.sage
   Description: Test file for redbivar.sage
   Copyright  : 2018, Robin Larrieu
-------------------------------------------------------------------------------
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
                   http://www.gnu.org/licenses/

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------
"""

load("redbivar.sage")
load("ggg.sage")

def hline():
    print "------------------------------------------------------"

def hline2():
    print "======================================================"

def result(b):
    print "OK" if b else "KO"


#============================
# Usual reduction algorithm
#============================

def test_reduction(G,Q,Rem,P):
    """
    Verify that the extended reduction is correct, that is
    P = sum_i Q_i * G_i + Rem
    """
    check = Rem
    for i in range(len(G)):
        check += Q[i] * G[i]
    return P == check
    
def test_reduction_zealous(G,P):
    """
    Verify correctness of the schoolbook reduction algorithm
    """
    print "Try to reduce P mod G (schoolbook algorithm)...",
    (Q,Rem) = dichotomic_zealous_reduction(G,P)
    result(test_reduction(G,Q,Rem,P))

def test_reduction_relaxed(G,P,k):
    """
    Verify correctness of the relaxed reduction algorithm
    (assumes that the term order being used is the k-order)
    """
    print "Try to reduce P mod G (relaxed algorithm)...",
    (Q,Rem) = dichotomic_relaxed_reduction(G,P,k)
    result(test_reduction(G,Q,Rem,P))


#============================
# Vanilla Groebner bases:
#   Generalities
#============================

def test_terse_repr(G, Gsharp, retrac, k):
    """
    Verify the correctness of the computed terse representation, that is 
    the retraction coefficients allow to reconstruct G from (G_0,G_1,G_n)
    """
    print "Check that the terse representation is correct:"
    print "Try to compute G_i using the retraction coefficients..."
    checkG = [ 0 for g in G ]
    n = len(G) - 1
    for i in [0,1,n]:
        checkG[i] = Gsharp[i]
    logn = len(retrac) # ceil(log_2 n)
    h = logn - 1
    l = 2^h
    while h >= 0:
        Ilc = I_lc(l,n)  # J_lambda \ J_{lambda+1}
        I2l = I_l(2*l,n) # J_{lambda+1}
        i = 0
        for C_lambda_i in retrac[h]:
            index_i = Ilc[i]
            j = 0
            for C_lambda_i_j in C_lambda_i:
                checkG[index_i] += C_lambda_i_j * checkG[ I2l[j] ]
                j += 1
            if G[index_i] != checkG[index_i]:
                print "  Computation of G_%d (l=%d) failed" % (index_i, l)
                print "KO"
                return false
            else:
                print "  Computation of G_%d (l=%d) OK" % (index_i, l)
            i += 1
        h -= 1
        l = l // 2
    print "OK"
    return true
    
def test_is_vanilla(p,q,k, strong_check=False):
    """
    Check that the Groebner basis for <p,q> w.r.t. the k-order is vanilla.
    By default, check for the retractive property for step lengths that are a power of 2;
    with strong_check=True, check for all step length.
    """
    #show_support(p) 
    G = bivar_groebner_basis(p,q,k_order(k))
    if not is_vanilla_stairs(G,k):
        print "Leading terms not in a vanilla Groebner stairs",
        return False
    l=2
    while l < len(G):
        try:
            invM = check_retractive(G,k,l) # raises an error if it is not retractive
        except VanillaGBError as e:
            print e,
            return False
        if strong_check:
            l += 1
        else:
            l *= 2
    return True


def test_vanilla_total_deg(d, strong_check=False):
    """
    Let I=<p,q> where p,q are random polynomials of total degree d.
    Check for various orders that the corresponding Groebner basis is vanilla.
    """
    p = R.random_element(degree = d, terms=Infinity)
    q = R.random_element(degree = d, terms=Infinity)
    print "Check for generators with total degree", d, "that the Groebner bases are vanilla"
    print "Degree of the ideal is", d*d
    k = 2
    while k < 2*d*d:
        print "  Test for the %d-order:" % k,
        sys.stdout.flush()
        if test_is_vanilla(p,q,k, strong_check):
            print "OK"
        else:
            print "KO"
            return
        k = 2*k

def test_vanilla_bivar_deg(d, strong_check=False):
    """
    Let I=<p,q> where p,q are random polynomials of degree d in both variables.
    Check for various orders that the corresponding Groebner basis is vanilla
    """
    p = S.random_element(degree = d, terms=Infinity)(x)
    q = S.random_element(degree = d, terms=Infinity)(x)
    for dy in range(d):
        p = y * p + S.random_element(degree = d, terms=Infinity)(x)
        q = y * q + S.random_element(degree = d, terms=Infinity)(x)
    print "Check for generators with degree", d, "in both variables that the Groebner bases are vanilla"
    print "Degree of the ideal is", 2*d*d
    k = 2
    while k < 4*d*d:
        print "  Test for the %d-order:" % k,
        sys.stdout.flush()
        if test_is_vanilla(p,q,k, strong_check):
            print "OK"
        else:
            print "KO"
            return
        k = 2*k

def test_vanilla_univ(d, strong_check=False):
    """
    Let I=<p,y + q> where p,q are random univariate polynomials of degree d.
    Check for various orders that the corresponding Groebner basis is vanilla
    """
    p = S.random_element(degree = d, terms=Infinity)(x)
    q = y + S.random_element(degree = d, terms=Infinity)(x)
    print "\nCheck for generators of the form p(x),y + q(x) (with p,q univariate of degree", d,
    "that the Groebner bases are vanilla"
    print "Degree of the ideal is", d
    k = 1
    while k < 2*d:
        print "  Test for the %d-order:" % k,
        sys.stdout.flush()
        if test_is_vanilla(p,q,k, strong_check):
            print "OK"
        else:
            print "KO"
            return
        k = 2*k


#============================
# Vanilla Groebner bases:
#   Reduction
#============================

def test_retract_quotients(G, retrac, k, d):
    """
    Verify the correctness of retracted quotients for random quotients of degree d
    (assuming the given retraction coefficients are correct).
    """
    print "Test algorithm for retraction of quotients (d=%d)..." % d,
    n = len(G) - 1
    Q = [ random_k_degree(k, d) for i in range(n+1) ]
    S = retract_quotients(Q, retrac)
    SG = S[0] * G[0] + S[1] * G[1] + S[2] * G[n]
    QG = 0
    for i in range(n+1):
        QG += Q[i] * G[i]
    result(SG == QG)

def test_vanilla_fast_reduce(G, TerseRepr, P, k, relaxed_algorithm=True):
    """
    Verify the correctness of the fast reduction algorithm (vanilla Groebner bases)
    """
    print "Reducing P mod G (deg P = %d)..." % P.degree(),
    (Q,Rem) = vanilla_fast_reduce(TerseRepr,P,k,relaxed_algorithm)
    result(test_reduction(G,Q,Rem,P))


#============================
# Main tests
#============================

def par(s):
    print
    hline2()
    print s
    hline2()
    print

def main_tests():
    k = 4
    print "Working on", R
    print "Run tests with k =", k
    #========================
    par("Test classical reduction algorithms")
    print "(I=<a,b> with a,b of total degree 10)"
    a = random_element(10)
    b = random_element(10)
    G = bivar_groebner_basis(a,b,k_order(k))
    print "-- P of total degree 10"
    P = random_element(10,k_order(k))
    test_reduction_zealous(G,P)
    test_reduction_relaxed(G,P,k)
    print "-- P of %d-degree 20" % k
    P = random_k_degree(k,20)
    test_reduction_zealous(G,P)
    test_reduction_relaxed(G,P,4)
    print "-- P of %d-degree 50" % k
    P = random_k_degree(k,50)
    test_reduction_zealous(G,P)
    test_reduction_relaxed(G,P,k)

    #========================
    par("Vanilla Groebner bases")
    hline()
    print "Construction of vanilla Groebner bases\n"
    test_vanilla_total_deg(10)
    test_vanilla_total_deg(20)
    test_vanilla_bivar_deg(10)
    test_vanilla_bivar_deg(20)
    test_vanilla_univ(100)
    test_vanilla_univ(500)

    print
    hline()
    print "Terse representation and reduction\n"
    print "Computing the terse representation of G..."
    TerseRepr = terse_repr(G,k)
    (Gsharp, retrac) = TerseRepr
    print "done\n"
    test_terse_repr(G, Gsharp, retrac, k)
    print
    test_retract_quotients(G, retrac, k, 10)
    test_retract_quotients(G, retrac, k, 30)
    print
    #reduction is faster with zealous reduction algorithm as backend
    test_vanilla_fast_reduce(G, TerseRepr, random_k_degree(k,20), k, relaxed_algorithm=False)
    test_vanilla_fast_reduce(G, TerseRepr, random_k_degree(k,30), k, relaxed_algorithm=False)
    test_vanilla_fast_reduce(G, TerseRepr, random_k_degree(k,40), k, relaxed_algorithm=False)
    test_vanilla_fast_reduce(G, TerseRepr, random_k_degree(k,50), k, relaxed_algorithm=False)

main_tests()

