"""
-------------------------------------------------------------------------------
   File       : redbivar.sage
   Description: A bivariate reduction algorithm for vanilla Groebner bases
   Copyright  : 2018, Robin Larrieu
-------------------------------------------------------------------------------
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
                   http://www.gnu.org/licenses/

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
-------------------------------------------------------------------------------

This is a proof-of-concept implementation of the fast reduction algorithm from:
  J. van der Hoeven and R. Larrieu.
  Fast reduction of bivariate polynomials with respect to sufficiently regular Groebner bases.
  Technical Report, HAL, 2018. http://hal.archives-ouvertes.fr/hal-01702547.

This algorithm is designed for so-called "vanilla Groebner bases", i.e. bases G
that verify the following properties:
  - The considered polynomial ring has 2 variables X,Y.
  - The monomials in normal form with respect to G are the minimal monomials for
    the relevant term order.
    Currently, only (1,k) weighted deglex orders (with k an integer) are considered.
  - Suitable relations exist among the basis elements.
See the above paper for more details

NOTE: some of the required primitives (e.g. relaxed multiplication) are 
implemented naively, so the current implementation does not achieve quasi-optimal
complexity as announced (see file bivariate.sage).
This implementation is mainly intended as a correctness proof of our algorithm.
"""

load("bivariate.sage")

class VanillaGBError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

#=============================
# Vanilla Groebner stairs
#=============================

def vanilla_stairs(D, k):
    """
    The leading terms for a Vanilla Groebner stair of an ideal of degree D, with respect to the k-order
    """
    n = Integer( ceil((sqrt(8*D/k+1) -1)/2) )
    u = D - k * (n*(n-1)) // 2
    q = u // n
    r = u  % n
    result = [y^n]
    for i in range(1,r):
        result.append(x^(q+k*(i-1)+1)*y^(n-i))
    if k > 1 and r > 0:
        result.append(x^(q+k*(r-1)+1)*y^(n-r))
    for i in range(r+1,n+1):
        result.append(x^(q+k*(i-1))*y^(n-i))
    return result

def is_vanilla_stairs(G, k):
    """
    Tests whether the leading terms of G follow a vanilla stairs
    """
    # Compute the degree of the ideal assuming G is vanilla
    D = 0
    LT = [ g.lt() for g in G ]
    for t in LT:
        if k == 1:
            D += t.degree(y)
        else:
            D += t.degree(x)
    # compare the leading terms with an actual vanilla stairs
    return LT == vanilla_stairs(D, k)


#=============================
# Remarkable sets of indices
#=============================

def I_l(l,n):
    """
    The set I_l = {0,1,n} U {l, 2l, ... , (n quo l)*l}.
    In the paper, we also use the shortcut J_lambda = I_{2^lambda}
    """
    Il = [0,1]
    i = l;
    if l == 1:
        i = 2
    while i < n:
        Il.append(i)
        i += l
    Il.append(n)
    return Il

def I_lc(l,n):
    """
    The set (I_l - I_2l)
    """
    i = l
    if l == 1:
        i = 3
    Ilc = []
    while i < n:
       Ilc.append(i)
       i += 2*l
    return Ilc

def I_2l_index(l,n):
    """
    The set of indices S such that I_l(2*l,n) = [ I_l(l,n)[s] for s in S ]
    """
    result = [0,1]
    s = 2 if l == 1 else 3
    Il = I_l(l,n)
    while s < len(Il) - 1:
        result.append(s)
        s += 2
    result.append(len(Il)-1)
    return result

def I_lc_index(l,n):
    """
    The set of indices S such that I_lc(l,n) = [ I_l(l,n)[s] for s in S ]
    """
    result = []
    s = 3 if l == 1 else 2
    Il = I_l(l,n)
    while s < len(Il) - 1:
        result.append(s)
        s += 2
    return result


#=============================
# Retraction coefficients
#=============================

def retraction_degree_bound(k,l):
    """
    The bound on the degree for the retraction coefficients
    """
    return k*(2*l-1)-1

def retraction_monomials(k,l):
    """
    The list of monomials with k-degree at most the allowed bound
    """
    d = retraction_degree_bound(k,l)
    result = []
    ypow = 0
    while k*ypow <= d:
        for xpow in range(d+1 - k*ypow):
            result.append(x^xpow * y^ypow)
        ypow += 1
    return result

def retraction_monomials_count(k,l):
    """
    The number of monomials of k-degree at most the allowed bound
    """
    ypow=0
    res=0
    d = retraction_degree_bound(k,l)
    while k*ypow <= d:
        res += d+1 - (k*ypow)
        ypow += 1
    return res

def image_basis(G,k,l):
    """
    The set of monomials {X^a*Y^b = lm(G_i)*m for some index i<n, and some monomial m with appropriate k_degree.
    This set is ordered such that the lm(G_i) come first for convenience.
    """
    n = len(G)-1
    LM = [ g.lm() for g in G ]
    monomials = retraction_monomials(k,l)
    i = 0
    result = copy(LM)
    for lm in LM:
        for a in range(1,len(monomials)): # discard the first element since 1*lm is already in result 
            m = lm * monomials[a]
            if (i>0) and LM[i-1].divides(m): break # all further elements are already in result
            result.append(m)
        i += 1
    return result

def matrix_of_retraction(G,k,l):
    """
    Assuming G is a vanilla Groebner basis w.r.t. the k-order and l=2^lambda, returns the matrix
    (in a well-chosen basis) of the linear map
      - (c_{j}) -> sum_j c_{j}*G_j
      - j in I_l, and deg_k c_{j} <= retraction_degree_bound(k,l)
    The source basis is the natural monomial basis for a tuple of polynomials, the image basis is as above.
    """
    n = len(G)-1
    LM = [ g.lm() for g in G ]
    indices = I_l(l,n)
    d = retraction_degree_bound(k,l)
    
    monomials = retraction_monomials(k,l)
    nbcol = len(indices) * len(monomials)
    
    img_basis = image_basis(G,k,l)
    nbrow = len(img_basis)
    M = Matrix(K,nbrow,nbcol)
    # fill the matrix column by column
    col=0
    for j in indices:
        Gj = G[j]
        for m in monomials:
            Gjm = Gj * m
            # fill the column corresponding to Gjm
            row = 0
            # The lm(G_i)*x^a for i <= n, a<d
            for img in img_basis:
                dy = img.degree(y)
                dx = img.degree(x)
                M[row,col] = Gjm[dx,dy]
                row += 1
            col += 1
    return M



def check_retractive(G,k,l, indices=[]):
    """
    Check that G is retractive for step length l; that is there are linear
    combinations
      - G_i = sum_{j in I_l} c_{i,j}*G_j
      - deg_k c_{i,j} <= retraction_degree_bound(k,l)
    and raise an exception otherwise.
    If a set of indices is given, check only for i in this set (for all i by default).
    Output a matrix that allows to reconstruct these combinations.
    """
    n = len(G)-1
    if n <= 2: # nothing to do
        return identity_matrix(R.base_ring(),n+1)
    M = matrix_of_retraction(G,k,l)
    invM = M.pseudoinverse()
    check = M * invM
    (nrow, ncol) = M.dimensions()
    # check that G_i is in the image of the linear map for all i
    check_indices = range(n+1) if indices == [] else indices
    for i in check_indices:
        elem = [ 0 for a in range(nrow) ]
        elem[i] = 1
        test = [ c for c in check.column(i) ]
        if elem != test:
            print elem
            print test
            raise VanillaGBError("Failed to find retraction coefficients for G_%d (l=%d)" % (i, 2*l))
    return invM


def vector_to_retrac_coef(V,indices,monomials):
    """
    Translate the vector V into a sequence of retraction coefficients (C_j), j in indices
    """
    # Sanity check
    if len(V) != len(monomials)*len(indices):
        raise VanillaGBError("Unexpected dimension: expected %d, got %d" % ( len(monomials)*len(indices), len(V) ))
    i = 0
    result = [ ]
    for j in indices:
        Cj = 0
        for m in monomials:
            Cj += V[i] * m
            i += 1
        result.append(Cj)
    return result



def retraction_coefficients(G,k,l):
    """
    Assuming G is a vanilla Groebner basis w.r.t. the k-order and l=2^lambda, compute the retraction
    coefficients C_{lambda,i,j} for i in (J_lambda - J_{lambda+1}), j in J_{lambda+1} such that
      - G_{i*l} = sum_j C_{lambda,i,j}*G_j
      - deg_k C_{lambda,i,j} <= retraction_degree_bound(k,2l)
    """
    n = len(G)-1
    result = [ ]
    I = I_lc(l,n)
    J = I_l(2*l,n)
    invM = check_retractive(G,k,2*l,I)
    for i in I:
        retrac = vector_to_retrac_coef(invM.column(i),J,retraction_monomials(k,2*l))
        result.append(retrac)
    return result


#=============================
# Terse representation
#=============================

def terse_repr(G,k):
    """
    The terse representation of G, that is the sequence G# of the G_i's truncated
    to the appropriate precision, together with the retraction coefficients
    """
    if not is_vanilla_stairs(G,k):
        raise VanillaGBError("Leading terms not in a vanilla Groebner stairs")
    Gsharp=[G[0],G[1]]
    n = len(G) - 1
    for i in range(2,n):
        p = 1
        while i % p == 0:
            p = 2*p
        p = k*p
        Gsharp.append(truncate_k_degree(G[i],k,p))
    Gsharp.append(G[n])
    l=1
    retrac = []
    while l < n:
        retrac.append(retraction_coefficients(G,k,l))
        l = 2*l
    return (Gsharp, retrac)


        
#===================================
# Fast reduction algorithm
#===================================

def retract_quotients(Q, retrac):
    """
    Given quotients Q=(Q_0,...,Q_n) and the retraction coefficients for G
    (as in the terse representation), compute retracted quotients S_0,S_1,S_n
    such that
      - S_0*G_0 + S_1*G_1 + S_n*G_n = Q_0*G_0 + ... + Q_n*G_n
    (algorithm 1 in the paper)
    """
    n = len(Q) - 1
    Q_lambda = Q
    l = 1
    for C_lambda in retrac:
        Ilc = I_lc(l,n)  # J_lambda \ J_{lambda+1}
        I2l = I_l(2*l,n) # J_{lambda+1}
        Ilc_index = I_lc_index(l,n)
        I2l_index = I_2l_index(l,n)
        new_Q_lambda = [ Q_lambda[i] for i in I2l_index ]
        j = 0
        for C_lambda_j in C_lambda: # j in Ilc = J_lambda \ J_{lambda+1}
            i = 0
            index_j = Ilc_index[j]
            for C_lambda_j_i in C_lambda_j: # i in I2l = J_{lambda+1}
                new_Q_lambda[ i ] += C_lambda_j_i * Q_lambda[ index_j ]
                i += 1
            j += 1
        Q_lambda = new_Q_lambda
        l *= 2
    if len(Q_lambda) != 3: # sanity check
        raise VanillaGBError("retract_quotients failed: expected 3 retracted quotients, got %d" % len(Q_lambda))
    return Q_lambda

  
def vanilla_fast_reduce(TerseRepr, P, k, relaxed_algorithm=True):
    """
    The extended reduction of P modulo G with respect to the k-order,
    where TerseRepr is the terse representation of G.
    Return (Q,Rem) with Q = [Q_0,...,Q_n] such that
      - P = Q_0*G_0 + ... + Q_n * G_n + Rem
      - No term in Rem is divisible by the leading term of any G_i
    The reduction is computed using the dichotomic selection strategy
    and the fast reduction algorithm (algorithm 2 in the paper)
    """
    (Gsharp, retrac) = TerseRepr
    if relaxed_algorithm:
        (Q, RemSharp) = dichotomic_relaxed_reduction(Gsharp, P, k)
    else:
        (Q, RemSharp) = dichotomic_zealous_reduction(Gsharp, force_term_order(P,k_order(k)))
    S = retract_quotients(Q, retrac)
    n = len(Gsharp) - 1
    Rem = P - S[0]*Gsharp[0] - S[1]*Gsharp[1] - S[2]*Gsharp[n]
    return (Q,Rem)

